from collections.abc import Mapping
from queue import Empty
import sys
from collections import namedtuple
from logging import getLogger, Logger, DEBUG
from time import monotonic_ns
from multiprocessing.dummy import Process, Queue
from typing import List, Dict
from datetime import datetime

import numpy as np

from .indexes import IndexUnderTest
from .dataset import BaseDataset
from .result import BenchmarkResult

SingleResult = namedtuple(
    'SingleResult', ['distances', 'labels', 'time', 'test_indexes', 'count']
)

class RunnerLog(Logger):
    def __init__(self, name, rlog):
        super().__init__(name, DEBUG)
        self._logger = getLogger('annb')
        self.rlog = rlog

    def _log(self, level, msg, *args, **kwargs):
        self._logger._log(level, msg, *args, **kwargs)
        if self.rlog:
            self.rlog._log(level, msg, *args, **kwargs)

class Runner:
    def __init__(self, name, index: IndexUnderTest, dataset: BaseDataset, **kwargs):
        self.name = name
        self.index = index
        self.dataset = dataset
        self.query_args = kwargs.get('query_args', [])
        self.topk = kwargs.get('topk', 10)
        self.step = kwargs.get('step', 10)
        self.jobs = kwargs.get('jobs', 1)
        self.loop = kwargs.get('loop', 5)
        self.benchmark_result = BenchmarkResult()
        self.loop_index = 0
        self.queue = Queue()
        self.records = {}
        for key, value in kwargs.items():
            self.benchmark_result.add_attribute(key, value)
        self.benchmark_result.add_attribute('name', self.name)
        self.benchmark_result.add_attribute('topk', self.topk)
        self.benchmark_result.add_attribute('step', self.step)
        self.benchmark_result.add_attribute('jobs', self.jobs)
        self.benchmark_result.add_attribute('loop', self.loop)
        self.benchmark_result.add_attribute('query_args', self.query_args)
        self.benchmark_result.add_attribute('dataset', self.dataset.name)
        self.benchmark_result.add_attribute('index', self.index.name)
        self.benchmark_result.add_attribute('dim', self.index.dimension)
        self.benchmark_result.add_attribute('metric_type', self.index.metric_type)
        self.benchmark_result.add_attribute('index_args', self.index.kwargs)
        self.benchmark_result.add_attribute(
            'started', datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        )
        if self.step == 0:
            # use all test data query once for batch mode
            self.step = self.dataset.test.shape[0]
        self.rlog = kwargs.get('rlog', None)
        self.log = RunnerLog(name, self.rlog)
        self.log.info('Runner init with %s', kwargs)

    def duration_run(self, text, func, *args, **kwargs):
        started = monotonic_ns()
        res = func(*args, **kwargs)
        duration = monotonic_ns() - started
        self.log.info('%s: %fms', text, duration / 1000000.0)
        return res, duration

    def run(self):
        self.index.cleanup()
        _, duration = self.duration_run(
            f'train {len(self.dataset.train)} items',
            self.index.train,
            self.dataset.train,
        )
        self.benchmark_result.add_training_duration(len(self.dataset.train), duration)
        _, duration = self.duration_run(
            f'add {len(self.dataset.data)} items', self.index.add, self.dataset.data
        )
        self.benchmark_result.add_insert_duration(len(self.dataset.data), duration)
        self.run_search_loop()

    def run_search_loop(self):
        query_args = self.query_args or [None]
        for i, query_arg in enumerate(query_args):
            if query_arg:
                if isinstance(query_arg, Dict):
                    self.index.update_search_args(**query_arg)
                    self.log.info('Update query args: %s', query_arg)
            self.records.clear()
            for loop_index in range(self.loop):
                self.loop_index = loop_index
                self.run_search()
            self.finalize_result(query_arg)
            self.log.info('Finish query args(%d/%d)', i + 1, len(query_args))

    @classmethod
    def run_multi_search(cls, args):
        for arg in args:
            cls.run_single_search(*arg)

    @classmethod
    def run_single_search(
        cls,
        index: IndexUnderTest,
        xq: np.array,
        test_indexes: List[int],
        topk: int,
        queue: Queue,
    ):
        assert len(xq) == len(test_indexes)
        start = monotonic_ns()
        distances, labels = index.search(xq, topk)
        end = monotonic_ns()
        result = SingleResult(distances, labels, end - start, test_indexes, len(xq))
        queue.put(result)

    def finalize_result(self, query_arg: Dict):
        best_loop = self.find_best_loop()
        best_results = self.records[best_loop]
        best_results = sorted(best_results, key=lambda r: r.test_indexes[0])
        np.concatenate([r.distances for r in best_results])
        labels = np.concatenate([r.labels for r in best_results])
        durations = [(len(r.test_indexes), r.time) for r in best_results]
        correct_count = 0
        total_count = 0
        ground_truth_neighbors = self.dataset.ground_truth_neighbors[:, : self.topk]

        # calc recall between ground_truth_neighbors and labels
        for i, (gt, test_items) in enumerate(zip(ground_truth_neighbors, labels)):
            total_count += len(gt)
            correct_count += len(set(gt) & set(test_items))
        recall = correct_count / total_count
        self.log.info('recall %.6f(%d/%d)', recall, correct_count, total_count)
        self.benchmark_result.add_query_result(
            recall=recall, durations=durations, query_arg=query_arg
        )

    def find_best_loop(self):
        # select which loop is the best
        best_loop = -1
        best_time = sys.maxsize
        for loop_index, records in self.records.items():
            time = 0
            indexes = []
            for record in records:
                time += record.time
                indexes.extend(record.test_indexes)
            indexes = sorted(indexes)
            if time < best_time and indexes == list(range(len(indexes))):
                best_loop = loop_index
                best_time = time
        if best_loop < 0:
            raise RuntimeError('No best loop found')
        self.log.info(
            '%s best loop: %d, with total duration: %fms',
            self.name,
            best_loop + 1,
            best_time / 1000000,
        )
        return best_loop

    def handle_result(self, result, proceed, total):
        self.log.debug(
            '%s single result: %d queries (%d/%d), %fms',
            self.name,
            result.count,
            proceed,
            total,
            result.time / 1000000,
        )
        self.records.setdefault(self.loop_index, []).append(result)

    def run_search(self):
        xq = self.dataset.test
        total_count = len(xq)
        jobs_args_list = [[]] * self.jobs
        for i in range(0, total_count, self.step):
            index = i // self.step % self.jobs
            indexes = list(range(i, min(i + self.step, total_count)))
            job_arg = (
                self.index,
                xq[i : i + self.step],
                indexes,
                self.topk,
                self.queue,
            )
            jobs_args_list[index].append(job_arg)
        jobs = []
        for pargs in jobs_args_list:
            p = Process(target=self.run_multi_search, args=(pargs,))
            jobs.append(p)
            p.start()
        # collect the result
        for i in range(0, total_count, self.step):
            try:
                ret = self.queue.get(timeout=180)
            except Empty:
                self.log.error(
                    'Timeout when waiting for result, not result return in 180 seconds'
                )
                break
            if isinstance(ret, str):
                self.log.debug('debug from subprocess: %s', ret)
            else:
                proceed_count = i + self.step
                self.handle_result(ret, proceed_count, total_count)
        for p in jobs:
            p.join()
        self.log.info(
            'Finish %d queries in loop(%d/%d)',
            total_count,
            self.loop_index + 1,
            self.loop,
        )



from argparse import ArgumentParser
from typing import Union
from logging import getLogger, Formatter, StreamHandler, FileHandler
from sys import stdout
from tempfile import gettempdir
from os import path

from .dataset.hdf5_dataset import AnnbHdf5Dataset
from .dataset.random_dataset import RandomDataset
from .indexes import IndexUnderTestFactory
from .plot import plot_result_recall_vs_qps
from .result import BenchmarkResult
from .runner import Runner
from .config import load_configs
from . import __version__ as annb_version

logger = getLogger('annb')


def create_logger(name: str, log_level: Union[str, int], log_file: Union[str, None] = None):
    logger = getLogger(name)
    if isinstance(log_level, str):
        log_level = log_level.upper()
    handler = StreamHandler(stdout)
    if log_file:
        handler = FileHandler(log_file)
    handler.setLevel(log_level)
    handler.setFormatter(
        Formatter(
            '%(asctime)s [%(name)s][%(levelname)s] %(filename)s:%(lineno)d, %(message)s'
        )
    )
    logger.addHandler(handler)
    logger.setLevel(log_level)
    return logger


def init_logger(log_level: str, log_file: Union[str, None] = None):
    """Initialize logger"""
    return create_logger('annb', log_level, log_file)


def load_dict(s):
    """
    >>> load_dict('a=1,b=2')
    {'a': 1, 'b': 2}
    """
    s = s.strip()
    data = {}
    for x in s.split(','):
        k, v = x.split('=')
        possible_types = [int, float, str]
        for t in possible_types:
            try:
                v = t(v)
                break
            except ValueError:
                pass
        data[k] = v
    return data


def load_index_factory(index_factory, index_factory_args) -> IndexUnderTestFactory:
    """
    >>> load_index_factory('annb.anns.faiss.indexes.index_under_test_factory')
    <annb.anns.faiss.indexes.FaissIndexUnderTestFactory object at 0x7f6b3d0b9e10>
    """
    try:
        module_name, class_name = index_factory.rsplit('.', 1)
        module = __import__(module_name, fromlist=[class_name])
        return getattr(module, class_name)(**index_factory_args)
    except Exception as e:
        logger.error('Load index factory failed', e)
        exit(1)


def create_or_load_dataset(dataset_file: str, dimension: int, metric_type: str, count: int):
    """
    >>> create_or_load_dataset('sift-128-euclidean.hdf5', 128, 'euclidean')
    <annb.dataset.hdf5_dataset.AnnbHdf5Dataset object at 0x7f6b3d0b9e10>
    """
    if dataset_file:
        try:
            dataset = AnnbHdf5Dataset(dataset_file)
        except Exception as e:
            logger.error('Load dataset failed', e)
            exit(1)
    else:

        temp_file = path.join(gettempdir(), f'.annb_random_d{dimension}_{metric_type}_{count}.hdf5')
        dataset = RandomDataset(temp_file, dimension=dimension, metric=metric_type, count=count)
    logger.info(
        'use dataset: %s, dim: %d, metric: %s',
        dataset,
        dataset.dimension,
        dataset.metric_type,
    )
    return dataset, dataset.dimension, dataset.metric_type


def run_once(
    name,
    index_factory,
    index_factory_args,
    index_name,
    index_dim,
    index_metric_type,
    index_args,
    query_args,
    dataset,
    result,
    result_log,
    topk,
    jobs,
    loop,
    step,
    count,
):
    factory = load_index_factory(index_factory, index_factory_args)
    dataset, index_dim, index_metric_type = create_or_load_dataset(
        dataset, index_dim, index_metric_type, count
    )
    index = factory.create(index_name, index_dim, index_metric_type, **index_args)
    logger.info(
        'use index: %s, dim: %d, metric: %s, %s',
        index,
        index.dimension,
        index.metric_type,
        index_args,
    )
    rlog = None
    if result and result_log:
        result_log_file = result + '.log'
        rlog = create_logger('annb.run-' + name, logger.level, result_log_file)
    runner = Runner(
        name,
        index,
        dataset,
        query_args=query_args,
        topk=topk,
        jobs=jobs,
        loop=loop,
        step=step,
        rlog=rlog,
    )
    runner.run()
    if result:
        logger.info('save result to %s', result)
        runner.benchmark_result.save(result)
    else:
        print(runner.benchmark_result)


def run_file(filename):
    runs = load_configs(filename)
    for run in runs:
        run_once(
            run['name'],
            run['index_factory'],
            run['index_factory_args'],
            run['index_name'],
            run['index_dim'],
            run['index_metric_type'],
            run['index_args'],
            run['query_args'],
            run['dataset'],
            run['result'],
            run['result_log'],
            run['topk'],
            run['jobs'],
            run['loop'],
            run['step'],
            run.get('count', 1000),
        )


def report_plain(inputs, output):
    output_file = stdout
    if output:
        output_file = open(output, 'w')
    for input in inputs:
        result = BenchmarkResult.load(input)
        output_file.write(f'# result for {input}:\n')
        output_file.write(f'{result}\n')
    output_file.close()


def report_csv(inputs, output):
    output_file = stdout
    if output:
        output_file = open(output, 'w')
    output_file.write(','.join(BenchmarkResult.csv_header) + '\n')
    for input in inputs:
        result = BenchmarkResult.load(input)
        for line in result.csv_output_lines():
            output_file.write(','.join([f'"{x}"' for x in line]) + '\n')


def report_png(inputs, output, **kwargs):
    data = [BenchmarkResult.load(input) for input in inputs]
    plot_result_recall_vs_qps(data, output=output, **kwargs)


def pth_file_path(file_path: str):
    """
    >>> pth_file_path('a.pth')
    'a.pth'
    >>> pth_file_path('a')
    'a.pth'
    """
    if file_path and not file_path.endswith('.pth'):
        file_path = file_path + '.pth'
    return file_path


def report_main():
    parser = ArgumentParser()
    parser.add_argument('input', default=[], help='Input report files', nargs='+')
    parser.add_argument(
        '--format',
        default='plain',
        choices=['csv', 'plain', 'png'],
        help='Output format',
    )
    parser.add_argument(
        '--output', default='', help='Output pth file, if not set will print to stdout'
    )
    parser.add_argument(
        '--title',
        default='Recall vs QPS',
        help='Title for png report, only used when format is png',
    )
    parser.add_argument(
        '--subtitle',
        default='',
        help='Subtitle for png report, only used when format is png',
    )
    parser.add_argument(
        '--foot-notes',
        default='',
        help='Foot notes for png report, only used when format is png',
    )
    opts = parser.parse_args()
    if not opts.input:
        print('No input files provided')
        exit(1)
    if opts.format == 'plain':
        report_plain(opts.input, opts.output)
    elif opts.format == 'csv':
        report_csv(opts.input, opts.output)
    elif opts.format == 'png':
        report_png(
            opts.input,
            opts.output,
            title=opts.title,
            subtitle=opts.subtitle,
            foot_notes=opts.foot_notes,
        )


def test_main():
    parser = ArgumentParser()
    parser.add_argument(
        '--log-level',
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
        help='Log level',
    )
    parser.add_argument(
        '--log-file', default=None, help='Log file, if not specified, log to stdout'
    )
    parser.add_argument(
        '--result-log', default=False, help='Log each run result, only used for debug/develop',
         action='store_true'
    )
    parser.add_argument(
        '--run-file', default='', help='Run config file, if not set use config from cli'
    )
    # options for run
    parser.add_argument(
        '--name', default='Test', help='Run name, if not set use index name'
    )
    parser.add_argument(
        '--index-factory', default='annb.anns.faiss.indexes.index_under_test_factory'
    )
    parser.add_argument(
        '--index-factory-args',
        default={},
        type=load_dict,
        help='Index factory args, comma separated key=value',
    )
    parser.add_argument('--index-name', default='Test', help='Index name')
    parser.add_argument(
        '--index-dim',
        default=256,
        type=int,
        help='Index dimension, only used when generate random dataset',
    )
    parser.add_argument(
        '--index-metric-type',
        default='l2',
        help='Index metric type, only used when generate random dataset',
    )
    parser.add_argument(
        '--index-args',
        default={'index': 'ivfflat', 'nlist': 128},
        type=load_dict,
        help='Index args, comma separated key=value',
    )
    parser.add_argument(
        '--query-args',
        default=[{'nprobe': 1}],
        type=load_dict,
        action='append',
        help='Query args, comma separated key=value, set multiple times to run multiple queries',
    )
    parser.add_argument('--topk', default=10, type=int, help='topk')
    parser.add_argument(
        '--step', default=10, type=int, help='step size, also as batch size, if use 0, will query all test data once'
    )
    parser.add_argument(
        '--batch', default=False, action='store_true', help='batch mode, alias --step 0'
    )
    parser.add_argument(
        '--jobs',
        default=1,
        type=int,
        help='jobs, how many query jobs to run in parallel',
    )
    parser.add_argument(
        '--loop',
        default=5,
        type=int,
        help='loop, how many times to run the query, only use the best one',
    )
    parser.add_argument(
        '--dataset',
        default='',
        help='Dataset file, if not set will generate random dataset',
    )
    parser.add_argument(
        '--result', default='', type=pth_file_path, help='Result file, if not set will print to stdout'
    )
    parser.add_argument(
        '--count',
        default=1000,
        type=int,
        help='Count, only used when generate random dataset',
    )
    parser.add_argument(
        '--version', action='version', version='%(prog)s ' + annb_version
    )

    opts = parser.parse_args()
    init_logger(opts.log_level, opts.log_file)

    if opts.run_file:
        logger.debug('run with file: %s', opts.run_file)
        run_file(opts.run_file)
    else:
        if opts.batch:
            opts.step = 0
        logger.debug('run with options: %s', opts)
        run_once(
            opts.name,
            opts.index_factory,
            opts.index_factory_args,
            opts.index_name,
            opts.index_dim,
            opts.index_metric_type,
            opts.index_args,
            opts.query_args,
            opts.dataset,
            opts.result,
            opts.result_log,
            opts.topk,
            opts.jobs,
            opts.loop,
            opts.step,
            opts.count,
        )


if __name__ == '__main__':
    test_main()



from typing import List, Dict
import os
import pickle

from annb.indexes import MetricType


class DurationWithCount:
    def __init__(self, count: int, duration: int):
        self.count = count
        self.duration = duration


class QueryResult:
    def __init__(self, recall: float, durations: List[DurationWithCount], args):
        self.recall = recall
        self.durations = durations
        self.args = args


class BenchmarkResult:
    csv_header = (
        'Test Name',
        'Started',
        'Index Name',
        'Index Dim',
        'Index Metric Type',
        'Index Args',
        'Topk',
        'Step',
        'Jobs',
        'Loop',
        'Dataset',
        'Training Durations(ms)',
        'Insert Durations(ms)',
        'Query Args',
        'Recall',
        'QPS',
        'Query Latency(ms)',
        'Query Latency P95(ms)',
        'Query Latency P99(ms)',
    )

    def __init__(self):
        self.training_durations = []
        self.insert_durations = []
        self.query_results = []
        self.attributes = {}

    def add_training_duration(self, count, duration):
        self.training_durations.append(DurationWithCount(count, duration))

    def add_insert_duration(self, count, duration):
        self.insert_durations.append(DurationWithCount(count, duration))

    def add_query_result(self, recall, durations: List, query_arg: Dict):
        result = QueryResult(recall, [], query_arg)
        for count, duration in durations:
            result.durations.append(DurationWithCount(count, duration))
        self.query_results.append(result)

    def add_attribute(self, key, value):
        self.attributes[key] = value

    def csv_output_lines(self):
        for query_result in self.query_results:
            name = self.attributes.get('name', 'Test')
            started = self.attributes.get('started', 'NONE')
            index_name = self.attributes.get('index', 'Test')
            index_dim = self.attributes.get('dim', 0)
            index_metric_type = self.attributes.get('metric_type', 'Unknown')
            if isinstance(index_metric_type, MetricType):
                index_metric_type = index_metric_type.name
            index_args = self.attributes.get('index_args', {})
            topk = self.attributes.get('topk', 10)
            step = self.attributes.get('step', 10)
            jobs = self.attributes.get('jobs', 1)
            loop = self.attributes.get('loop', 5)
            dataset = self.attributes.get('dataset', 'Unknown')
            training_durations = sum([d.duration for d in self.training_durations])
            insert_durations = sum([d.duration for d in self.insert_durations])
            yield (
                name,
                started,
                index_name,
                str(index_dim),
                index_metric_type,
                str(index_args),
                str(topk),
                str(step),
                str(jobs),
                str(loop),
                dataset,
                str(training_durations / 1000000.0),
                str(insert_durations / 1000000.0),
                str(query_result.args),
                str(query_result.recall),
                str(BenchmarkResult.qps(query_result.durations)),
                str(BenchmarkResult.latency(query_result.durations) / 1000000.0),
                str(BenchmarkResult.latency_pn(query_result.durations, 95) / 1000000.0),
                str(BenchmarkResult.latency_pn(query_result.durations, 99) / 1000000.0),
            )

    @staticmethod
    def qps(durations):
        return sum([d.count for d in durations]) / (
            sum([d.duration for d in durations]) / 1000000000.0
        )

    @staticmethod
    def latency(durations):
        return sum([d.duration for d in durations]) / len(durations)

    @staticmethod
    def latency_pn(durations, n):
        duration_values = [d.duration for d in durations]
        duration_values.sort()
        return duration_values[int(len(duration_values) * n / 100)]

    @staticmethod
    def _summary(durations):
        return (
            f'{len(durations)} items,'
            f' {sum([d.count for d in durations])} total,'
            f' {sum([d.duration for d in durations])/1000000.0}ms'
        )

    @property
    def training_durations_summary(self):
        return self._summary(self.training_durations)

    @property
    def insert_durations_summary(self):
        return self._summary(self.insert_durations)

    def save(self, filename: str):
        abs_filepath = os.path.abspath(filename)
        os.makedirs(os.path.dirname(abs_filepath), exist_ok=True)
        with open(abs_filepath, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def load(filename: str):
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def __str__(self):
        attributes = ''
        for key, value in self.attributes.items():
            attributes += f'    {key}: {value}\n'
        query_durations = ''
        for query_result in self.query_results:
            args = ''
            if isinstance(query_result.args, Dict):
                for key, value in query_result.args.items():
                    args += f'{key}={value}'
            else:
                args = 'none'
            recall = f'recall={query_result.recall}'
            durations_total_query = sum([d.count for d in query_result.durations])
            durations_total_duration = sum([d.duration for d in query_result.durations])
            durations_total_qps = BenchmarkResult.qps(query_result.durations)
            latency = BenchmarkResult.latency(query_result.durations)
            latency_p95 = BenchmarkResult.latency_pn(query_result.durations, 95)
            latency_p99 = BenchmarkResult.latency_pn(query_result.durations, 99)
            duration_summary = f'{durations_total_query} items,'
            duration_summary += f' {durations_total_duration/1000000.0}ms,'
            duration_summary += f' {durations_total_qps}qps,'
            duration_summary += f' latency={latency/1000000.0}ms,'
            duration_summary += f' p95={latency_p95/1000000.0}ms,'
            duration_summary += f' p99={latency_p99/1000000.0}ms'
            query_durations += f'      {args},{recall} -> {duration_summary}\n'

        return f"""
BenchmarkResult:
  attributes:
{attributes}
  durations:
    training: {self.training_durations_summary}
    insert: {self.insert_durations_summary}
    query:
{query_durations}
"""



import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import Tuple, List, Union, Dict
from hashlib import sha1
from logging import getLogger

import numpy as np
from .envs import get_run_dir


class MetricType(Enum):
    INNER_PRODUCT = 1
    L2 = 2
    JACCARD = 3

    @classmethod
    def from_text(cls, text: str):
        if text.lower() == 'inner_product':
            return cls.INNER_PRODUCT
        elif text.lower() == 'ip':
            return cls.INNER_PRODUCT
        elif text.lower() == 'angular':
            return cls.INNER_PRODUCT
        elif text.lower() == 'euclidean':
            return cls.L2
        elif text.lower() == 'l2':
            return cls.L2
        elif text.lower() == 'jaccard':
            return cls.JACCARD
        else:
            raise ValueError('Unknown metric type: {}'.format(text))


class IndexUnderTest(ABC):
    """
    Abstract class for the index.
    """

    def __init__(
        self, index_name: str, dimension: int, metric_type: MetricType, **kwargs
    ):
        """
        :param index_name: Name of the index.
        :param dimension: Dimension of the index.
        :param metric_type: Metric type of the index.
        """
        self.name = index_name
        self.dimension = dimension
        self.metric_type = metric_type
        self.kwargs = kwargs
        self.log = getLogger('annb')

    def verify(self) -> bool:
        """
        Verify the index.
        :return: True if the index is valid, False otherwise.

        This method is used to verify that the index is valid.
        Index under test should be able to verify itself with small data set.
        """
        return self.metric_type in (MetricType.L2, MetricType.INNER_PRODUCT)

    @abstractmethod
    def cleanup(self) -> None:
        """
        Cleanup the index and related data.
        """
        pass

    @abstractmethod
    def train(self, data: np.ndarray) -> None:
        """
        Train the index
        :param data: List of data to train the index with.
        """
        pass

    def warmup(self) -> None:
        """
        Warmup the index, called before search.
        """
        pass

    @abstractmethod
    def add(self, data: np.ndarray) -> None:
        """
        Add data to the index.
        :param data: List of data to add to the index.
        """
        pass

    @abstractmethod
    def search(self, query: np.ndarray, k: int) -> Tuple[List[float], List[int]]:
        """
        Search the index.
        :param query: Query data.
        :param k: Number of nearest neighbors to return.
        :return: List of nearest neighbors, distances and ids
        """
        pass

    @abstractmethod
    def update_search_args(self, **kwargs) -> None:
        """
        Update search arguments.
        :param kwargs: Search arguments.
        """
        pass


class IndexUnderTestFactory(ABC):
    """
    Abstract class for the index factory.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @abstractmethod
    def create(
        self, index_name: str, dimension: int, metric_type: MetricType, **kwargs
    ) -> IndexUnderTest:
        """
        Create the index.
        :return: IndexUnderTest object.
        """
        pass


class IndexUnderTestDeployment:
    """
    Deployment for index
    """

    def deploy(self, **kwargs) -> Tuple[str, str]:
        """
        Deploy the index.
        :return: Tuple of deployment type and reference.

        if deployment type is builtin, reference is empty string.
            return ('builtin', '')
        if deployment type is docker, reference is docker image name.
            return ('docker', 'docker_image_name')
        if deployment type is use conda, reference is conda environment path.
            return ('conda', 'conda_env_path')
        if deployment type is use venv, reference is venv environment path.
            return ('venv', 'venv_path')
        """
        deployment_type = kwargs.get('deployment_type', 'builtin')
        if deployment_type == 'docker':
            if 'image' not in kwargs:
                raise ValueError('Docker image is not specified for docker deployment')
            if 'dockerfile' in kwargs:
                self.build_docker_image(kwargs['dockerfile'], kwargs['image'])
            return 'docker', kwargs['image']
        if deployment_type == 'venv':
            if 'requirements' not in kwargs:
                raise ValueError(
                    'Requirements file is not specified for venv deployment'
                )
            venv_path = self.create_venv(kwargs['requirements'])
            return 'venv', venv_path
        if deployment_type == 'conda':
            if 'environments' not in kwargs:
                raise ValueError(
                    'Environments file is not specified for conda deployment'
                )
            conda_env_path = self.create_conda_env(kwargs['environments'])
            return 'conda', conda_env_path
        return 'builtin', ''

    @classmethod
    def build_docker_image(cls, dockerfile: str, image: str) -> None:
        """
        Build docker image.
        :param dockerfile: Dockerfile to build docker image.
        :param image: Docker image name.
        """
        import docker

        client = docker.from_env()
        client.images.build(path=dockerfile, tag=image)

    @classmethod
    def create_venv(cls, requirements: Union[List[str], str]) -> str:
        import venv
        import subprocess

        # read requirements if it is a file
        if isinstance(requirements, str):
            if os.path.isfile(requirements):
                with open(requirements, 'r') as f:
                    requirements = f.read()
            requirements = requirements.split('\n')
        requirements = sorted(requirements)
        sha1sum = sha1()
        for text in requirements:
            sha1sum.update(text.encode('utf-8'))
        venv_path = os.path.join(get_run_dir(), 'venv-{}'.format(sha1sum.hexdigest()))
        if not os.path.isfile(os.path.join(venv_path, 'bin', 'python')):
            builder = venv.EnvBuilder(with_pip=True)
            builder.create(venv_path)
        requirements_path = os.path.join(venv_path, 'requirements.txt')
        with open(requirements_path, 'w') as f:
            f.write('\n'.join(requirements))
        # install
        ret = subprocess.check_call(
            [os.path.join(venv_path, 'bin', 'pip'), 'install', '-r', requirements_path]
        )
        if ret != 0:
            raise RuntimeError('Failed to install requirements')
        return venv_path

    @classmethod
    def create_conda_env(cls, environments: Union[Dict, str]) -> str:
        import subprocess
        import yaml

        if isinstance(environments, str):
            if os.path.isfile(environments):
                with open(environments, 'r') as f:
                    environments = f.read()
            environments = yaml.loads(environments, Loader=yaml.FullLoader)
        conda_name = environments['name']
        conda_path = os.path.join(get_run_dir(), 'conda-{}'.format(conda_name))
        conda_yaml_path = conda_path + '.yaml'
        with open(conda_yaml_path, 'w') as f:
            yaml.dump(environments, f)
        # create conda
        cls.get_conda_executable()
        ret = subprocess.check_call(
            ['conda', 'env', 'create', '-f', conda_yaml_path, '-p', conda_path]
        )
        if ret != 0:
            raise RuntimeError('Failed to create conda env')
        return conda_path

    @classmethod
    def get_conda_executable(cls) -> str:
        home = os.environ.get('HOME', '')
        if home:
            for conda_path in [
                os.path.join(home, 'miniconda3', 'bin', 'conda'),
                os.path.join(home, 'anaconda3', 'bin', 'conda'),
                os.path.join(home, 'anaconda', 'bin', 'conda'),
            ]:
                if os.path.isfile(conda_path):
                    return conda_path
        return 'conda'



from typing import List
import matplotlib.pyplot as plt

from .result import BenchmarkResult


def plot_result_recall_vs_qps(results: List[BenchmarkResult], **kwargs):
    output = kwargs.get('output', 'result.png')
    if not output:
        output = 'result.png'
    if not output.endswith('.png'):
        output += '.png'

    plt.style.use('seaborn-whitegrid')
    # create figure from results
    fig, ax = plt.subplots()

    title = kwargs.get('title', 'Recall vs QPS')
    subtitle = kwargs.get('subtitle', '')
    if subtitle:
        title += '\n' + subtitle
    ax.set_title(title)

    ax.set_xlabel('Recall')
    ax.set_ylabel('QPS')
    ax.set_xlim(0.0, 1.0)

    lines = []
    labels = []
    for result in results:
        recalls = [r.recall for r in result.query_results]
        qps = [BenchmarkResult.qps(r.durations) for r in result.query_results]
        lines.append(
            ax.plot(
                recalls, qps, linestyle='-', marker='o', label=result.attributes['name']
            )
        )
        labels.append(result.attributes['name'])
    foot_notes = kwargs.get('foot_notes', '')
    if foot_notes:
        ax.text(
            0.0,
            -0.1,
            foot_notes,
            horizontalalignment='left',
            verticalalignment='top',
            transform=ax.transAxes,
        )
    fig.legend(loc='right', bbox_to_anchor=(1.2, 0.5), ncol=1)
    fig.savefig(output, bbox_inches='tight', dpi=300)



from typing import Tuple

from annb.indexes import IndexUnderTestDeployment


class FaissIndexUnderTestDeployment(IndexUnderTestDeployment):
    def deploy(self, **kwargs) -> Tuple[str, str]:
        deployment_type = kwargs.get('deployment_type', 'builtin')
        if deployment_type == 'venv':
            kwargs['requirements'] = [
                'faiss-cpu', 'numpy'
            ]
        return super().deploy(**kwargs)


index_under_test_deployment = FaissIndexUnderTestDeployment



"""
run config file use yaml format:

---
default:
  index_factory: <default index factory, default annb.anns.faiss.indexes.index_under_test_factory>
  index_factory_args: <the default index factory args, if not set use {}>
  index_name: <the default index name, if not set use Test>
  index_dim: <the default index dimension, if not set use from dataset>
  index_metric_type: <the default index metric type, if not set use from dataset>
  index_args: <the default index args, if not set use {}>
  query_args: <the default query args, if not set use {}>
  topk: <the default topk, if not set use 10>
  step: <the default step, if not set use 10>
  jobs: <the default jobs, if not set use 1>
  loop: <the default loop, if not set use 5>
  dataset: <the default dataset, if not set use annb.RandomDataset>
  result: <the default result file, if not set use None>

runs:
  -  name: <the run name, if not set use index name>
     index_factory: <the index factory, if not set use default>
  
"""

import re
import yaml


def valiate_configs(configs, default_config):
    # handle result, we need to make sure not overrite the result file
    results = set()
    for config in configs:
        if "result" in config:
            if config["result"] in results:
                raise ValueError(f'result file {config["result"]} used more than once')
            results.add(config["result"])


def auto_process_result_param(config):
    """auto rename result in default config"""
    if "result" in config and config["result"]:
        result_file_path = config["result"]
        if not result_file_path.endswith(".pth"):
            result_file_path += ".pth"
        slice_index = 0
        if re.match(r".*-\d+\.pth", result_file_path):
            slice_index = int(result_file_path.split("-")[-1].split(".")[0])
            result_file_path = result_file_path.replace(
                f"-{slice_index}.pth", f"-{slice_index+1}.pth"
            )
            slice_index += 1
        else:
            slice_index += 1
            result_file_path = result_file_path.replace(".pth", f"-{slice_index}.pth")
        config["result"] = result_file_path


def load_configs(filename: str):
    default_config = {
        "name": "Test",
        "index_factory": "annb.anns.faiss.indexes.index_under_test_factory",
        "index_factory_args": {},
        "index_name": "Test",
        "index_dim": None,
        "index_metric_type": None,
        "index_args": {},
        "query_args": [[{"nprobe": 1}]],
        "topk": 10,
        "step": 10,
        "jobs": 1,
        "loop": 5,
        "dataset": "annb.RandomDataset",
        "result": None,
        "result_log": False,
    }
    data = {}
    with open(filename, "r") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
    if "default" in data:
        default_config.update(data["default"])
    runs = []
    for run in data.get("runs", []):
        auto_process_result_param(default_config)
        run_config = default_config.copy()
        run_config.update(run)
        runs.append(run_config)
    valiate_configs(runs, default_config)
    return runs



from abc import ABC, abstractmethod
import numpy as np

from ..indexes import MetricType


class BaseDataset(ABC):
    name = 'Dataset'
    metric_type = MetricType.L2
    dimension = 0
    count = 0

    def __init__(self, **kwargs):
        """
        :param dataset_name: Name of the dataset.
        """
        self.kwargs = kwargs

    @property
    @abstractmethod
    def data(self) -> np.ndarray:
        """
        Return the dataset to be added to the index.
        """
        pass

    @property
    @abstractmethod
    def train(self) -> np.ndarray:
        """
        Return dataset to be used for training.
        could use different dataset for training and adding.
        """
        pass

    @property
    @abstractmethod
    def test(self) -> np.ndarray:
        """
        Return dataset to be used for search testing.
        """
        pass

    @property
    @abstractmethod
    def ground_truth_neighbors(self) -> np.ndarray:
        """
        Return ground truth for dataset. ID labels of nearest neighbors.
        """
        pass

    @property
    @abstractmethod
    def ground_truth_distances(self) -> np.ndarray:
        """
        Return ground truth for dataset. Distances to nearest neighbors.
        """
        pass



from .base_dataset import BaseDataset
from .hdf5_dataset import AnnbHdf5Dataset, Hdf5Dataset
from .random_dataset import RandomDataset

__all__ = ['Hdf5Dataset', 'AnnbHdf5Dataset', 'BaseDataset', 'RandomDataset']



from os import path
from typing import Union
import h5py as h5
import numpy as np

from ..indexes import MetricType
from .base_dataset import BaseDataset
from .utils import generate_groundtruth


class Hdf5Dataset(BaseDataset):
    """
    Dataset from hdf5 file.
    This dataset is for load dataset from hdf5 file from ann-benchmarks.
    """

    def __init__(self, file: Union[str, h5.File], **kwargs):
        """
        :param file: Path or hdf5 file to the dataset file.
        """
        super().__init__(**kwargs)
        if isinstance(file, str):
            self.hd5_file = self.load_hdf5(file)
        elif isinstance(file, h5.File):
            self.hd5_file = file
        self.validate()
        if 'dimension' in self.hd5_file.attrs:
            self.dimension = self.hd5_file.attrs['dimension']
            self.count = self.data.shape[0] / self.dimension
        else:
            self.dimension = self.data.shape[1]
            self.count = self.data.shape[0]
        self.metric_type = MetricType.from_text(str(self.hd5_file.attrs['distance']))
        self.name = path.basename(self.hd5_file.filename)

    def __str__(self) -> str:
        return (
            f'<{self.__class__.__name__}({self.name}, m={self.metric_type.name},'
            f' d={self.dimension}, nb={self.count})>'
        )

    def validate(self):
        """
        Validate dataset.
        """
        if 'distance' not in self.hd5_file.attrs:
            raise RuntimeError('dataset file not found distance attribute.')
        distance_text = str(self.hd5_file.attrs['distance'])
        self.metric_type = MetricType.from_text(distance_text)

        for required_dataset in ('distances', 'neighbors', 'test', 'train'):
            if required_dataset not in self.hd5_file:
                raise RuntimeError(
                    f'dataset file not found {required_dataset} dataset.'
                )

    @classmethod
    def load_hdf5(cls, cache_file: str) -> h5.File:
        """
        Load dataset from cache file.
        """
        if not cache_file or not path.exists(cache_file):
            raise FileNotFoundError(f'file {cache_file} not found.')
        hd = h5.File(cache_file)
        return hd

    @classmethod
    def create(
        cls,
        output: str,
        metric: MetricType,
        data_and_train: np.ndarray,
        test: Union[np.ndarray, None] = None,
        distances: Union[np.ndarray, None] = None,
        neighbors: Union[np.ndarray, None] = None,
        normalize: bool = False,
        ground_truth: bool = True,
        extra_attrs: dict = None,
    ):
        """
        Create dataset and save to file
        :param output: Output file path.
        :param metric: Metric type.
        :param data_and_train: Data and train data.
        :param test: Test data.
        :param distances: Distances data.
        :param neighbors: Neighbors data.
        :param normalize: Normalize data.
        :param ground_truth: Generate ground truth.
        :param extra_attrs: Extra attributes.
        """

        def get_distance_text(the_metric: MetricType) -> str:
            return 'euclidean' if the_metric == MetricType.L2 else 'angular'

        if normalize and metric == MetricType.L2:
            raise ValueError('normalize only support angular/ip metric.')
        if normalize and (distances is not None or neighbors is not None):
            raise ValueError(
                'normalize is set, but distances/neighbors data is not None.'
            )

        hd = h5.File(output, 'w')
        hd.attrs['distance'] = get_distance_text(metric)
        extra_attrs = extra_attrs or {}
        for k, v in extra_attrs.items():
            hd.attrs[k] = v

        # do normalize for data and train
        if normalize:
            data_and_train /= np.linalg.norm(data_and_train, axis=1)[:, np.newaxis]
        hd.create_dataset('train', data=data_and_train)

        if test is None:
            test_size = min(data_and_train.shape[0], 10000)
            test = data_and_train[:test_size]
            neighbors, distances = None, None
        else:
            # normalize for test data
            test /= np.linalg.norm(test, axis=1)[:, np.newaxis]

        if neighbors is None or distances is None:
            if ground_truth:
                distances, neighbors = generate_groundtruth(
                    test, data_and_train, metric
                )
            else:
                # return forge data, if no ground truth is needed
                neighbors = np.zeros((test.shape[0], 1), dtype=np.int64)
                distances = np.zeros((test.shape[0], 1), dtype=np.float32)

        hd.create_dataset('test', data=test)
        hd.create_dataset('neighbors', data=neighbors)
        hd.create_dataset('distances', data=distances)
        hd.close()

    @property
    def data(self) -> np.ndarray:
        # using train dataset as data
        return np.array(self.hd5_file['train'])

    @property
    def train(self) -> np.ndarray:
        return np.array(self.hd5_file['train'])

    @property
    def test(self) -> np.ndarray:
        """
        Return query test dataset.
        """
        return np.array(self.hd5_file['test'])

    @property
    def ground_truth_distances(self) -> np.ndarray:
        """
        Return dataset data.
        """
        return np.array(self.hd5_file['distances'])

    @property
    def ground_truth_neighbors(self) -> np.ndarray:
        """
        Return dataset ground truth.
        """
        return np.array(self.hd5_file['neighbors'])


class AnnbHdf5Dataset(Hdf5Dataset):
    """
    Dataset from hdf5 file, with some extra attributes from annb.
    """

    def __init__(self, file: Union[str, h5.File], **kwargs):
        """
        :param file: Path or hdf5 file to the dataset file.
        """
        super().__init__(file, **kwargs)
        self.normalized = self.hd5_file.attrs.get('normalized', False)

    def __str__(self) -> str:
        return (
            f'<{self.__class__.__name__}({self.name},'
            f' m={self.metric_type.name}, d={self.dimension},'
            f' nb={self.count}, normalized={self.normalized})>'
        )

    @classmethod
    def create(
        cls,
        output: str,
        metric_type: MetricType,
        data_and_train: np.ndarray,
        test: Union[np.ndarray, None] = None,
        distances: Union[np.ndarray, None] = None,
        neighbors: Union[np.ndarray, None] = None,
        normalize: bool = False,
        ground_truth: bool = True,
        extra_attrs: dict = None,
    ):
        extra_attrs = extra_attrs or {}
        extra_attrs['normalized'] = normalize
        return Hdf5Dataset.create(
            output,
            metric_type,
            data_and_train,
            test,
            distances,
            neighbors,
            normalize,
            ground_truth,
            extra_attrs,
        )
